#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  5 16:29:58 2025
"""

import numpy as np
import pandas as pd

try:
    from scipy.stats import norm
except Exception:
    pass


def robust_pinv(M, rcond=1e-15, reg=1e-8):
    # Add a small regularization term to the diagonal
    M_reg = M + reg * np.eye(M.shape[0])
    return np.linalg.pinv(M_reg, rcond=rcond)


def kalman_fusion_forecasts(
    F: np.ndarray,
    index,
    columns,
    coverage: float = 0.90,
    method: str = "multi_series",
    Q_init=0.1,
    R_init=1.0,
    adapt_Q: str = None,  # "none" or "spread"
    adapt_R: str = None,  # "none" or "spread"
    initial_x=None,
    initial_P=None,
    min_std=1e-15,
    scale: bool = True,
    a: float = 1.0,  # autoregressive term
):
    """
    Fuse multiple forecasts using a Kalman Filter for each forecast step.

    Args:
        F (np.ndarray): Forecasts of shape (n, S, T)
            - n = number of models
            - S = forecast length (time steps)
            - T = number of time series
        index (pd.Index): Index for the returned DataFrames (length=S).
        columns (list-like): Column names for the returned DataFrames (length=T).
        coverage (float): e.g. 0.90 => ~1.645 standard deviations for intervals.
        method (str): Either "multi_series" or "per_series".
            - "multi_series": single (T-dim) Kalman Filter for all T series together
            - "per_series": run T separate 1D Kalman Filters, one per series
        Q_init (float or np.ndarray): Base process noise (or matrix).
        R_init (float or np.ndarray): Base measurement noise (or matrix).
        adapt_Q (str): If "spread", adapt Q each step based on model spread.
        adapt_R (str): If "spread", adapt R each step based on model spread.
        initial_x (np.ndarray): initial state guess.
            - For multi_series: shape (T,)
            - For per_series: shape (T,)
        initial_P (np.ndarray): initial covariance guess.
            - For multi_series: shape (T, T)
            - For per_series: shape (T,)
        min_std (float): Small floor value for numerical stability in standard deviations.
        scale (bool): If True, each series is scaled (by its standard deviation) before filtering.
                      The fused forecasts are converted back to the original scale.

    Returns:
        df_point (pd.DataFrame): Kalman-fused point forecasts, shape (S, T).
        df_lower (pd.DataFrame): Lower bound, shape (S, T).
        df_upper (pd.DataFrame): Upper bound, shape (S, T).

    Notes:
        - In "multi_series" mode the state is a T-dimensional vector that is updated with an (n*T)-dimensional
          measurement (the stacked forecasts from the n models).
        - In "per_series" mode the algorithm runs T separate 1D Kalman Filters (one per series). In that case,
          for each forecast step (loop over S) you update each of the T filters independently.
        - When scale=True, each series is divided by its standard deviation (computed over all forecasts)
          before filtering; after filtering the outputs are multiplied by the same scale so that the returned forecasts
          are in the original feature space.
    """
    # shapes
    n, S, T = F.shape

    # Optional scaling: compute per-series scales and rescale F (and initial values)
    if scale:
        # Compute a scale (here, standard deviation) for each time series (last axis)
        scales = np.std(F, axis=(0, 1), ddof=1)
        scales[scales == 0] = 1.0  # avoid division by zero

        # Scale the forecasts: broadcast scales to shape (1, 1, T)
        F = F / scales[None, None, :]

        # Adjust initial state (if provided) to be in the scaled space.
        if initial_x is not None:
            initial_x = np.asarray(initial_x, dtype=float) / scales

        # Adjust the initial covariance.
        if initial_P is not None:
            if method == "multi_series":
                initial_P = np.array(initial_P, dtype=float)
                # If initial_P is a full matrix, scale it as: P_scaled = D^-1 P D^-1
                D_inv = np.diag(1.0 / scales)
                initial_P = D_inv @ initial_P @ D_inv
            else:  # per_series mode: initial_P is a vector
                initial_P = np.array(initial_P, dtype=float) / (scales**2)

    # z-value for coverage (assuming normal)
    z = norm.ppf(0.5 + coverage / 2.0)

    # -------------
    # Helper: build Q, R each step, possibly adapting
    # -------------
    def compute_process_noise(k, base_Q, x_dim, spread_value=1.0):
        """
        Return Q_k for time step k, shape = (x_dim, x_dim).
        If adapt_Q=="spread", scale by spread_value.
        """
        if np.isscalar(base_Q):
            val = base_Q
            if adapt_Q == "spread":
                val *= spread_value  # scale
            return val * np.eye(x_dim)
        else:
            # base_Q is a matrix
            if adapt_Q == "spread":
                return base_Q * spread_value
            else:
                return base_Q

    def compute_measurement_noise(k, base_R, meas_dim, spread_value=1.0):
        """
        Return R_k for time step k, shape = (meas_dim, meas_dim).
        If adapt_R=="spread", scale by spread_value.
        """
        if np.isscalar(base_R):
            val = base_R
            if adapt_R == "spread":
                val *= spread_value
            return val * np.eye(meas_dim)
        else:
            # base_R is a matrix
            if adapt_R == "spread":
                return base_R * spread_value
            else:
                return base_R

    # -------------
    # MODE 1: MULTI-SERIES (method="multi_series")
    # A single T-dimensional Kalman filter is used; the measurement at each step
    # is an (n*T)-dimensional vector formed by stacking all n models’ forecasts.
    # -------------
    if method == "multi_series":
        # State dimension = T
        x_dim = T
        # Measurement dimension = n*T
        meas_dim = n * T

        # Build measurement matrix H of shape (n*T, T) by stacking n identity matrices.
        H = np.tile(np.eye(T), (n, 1))  # shape = (n*T, T)

        # Transition matrix A = I (random walk)
        A = a * np.eye(T)
        # it woulds be useful to add a local linear trend version as an option here

        if initial_x is None:
            x_0 = np.mean(F[:, 0, :], axis=0)  # shape (T,)
        else:
            x_0 = np.asarray(initial_x, dtype=float).reshape((T,))

        # Set up initial covariance P_0
        if initial_P is None:
            P_0 = np.eye(T)
        else:
            P_0 = np.array(initial_P, dtype=float)
            if P_0.shape != (T, T):
                raise ValueError("initial_P must be shape (T,T) for multi_series mode")

        # Prepare arrays for outputs
        X_kalman = np.zeros((S, T))
        X_lower = np.zeros((S, T))
        X_upper = np.zeros((S, T))

        # Initialize current state and covariance
        x_curr = x_0
        P_curr = P_0

        for k in range(S):
            # Compute spread value if adapting Q or R:
            if adapt_R == "spread" or adapt_Q == "spread":
                # F[:, k, :] has shape (n, T); compute variance per series and average them.
                var_per_series = np.var(F[:, k, :], axis=0, ddof=1)
                spread_value = float(np.mean(var_per_series))
            else:
                spread_value = 1.0

            # Time update (prediction)
            x_pred = A @ x_curr
            Q_k = compute_process_noise(k, Q_init, x_dim, spread_value=spread_value)
            P_pred = A @ P_curr @ A.T + Q_k

            # Measurement update
            z_k = F[:, k, :].reshape(n * T)  # stack the forecasts into a (n*T,) vector
            y_k = H @ x_pred
            R_k = compute_measurement_noise(
                k, R_init, meas_dim, spread_value=spread_value
            )
            S_k = H @ P_pred @ H.T + R_k
            K_k = P_pred @ H.T @ np.linalg.pinv(S_k)
            x_update = x_pred + K_k @ (z_k - y_k)
            I_t = np.eye(T)
            tmp = I_t - K_k @ H
            # Joseph form for stability:
            P_update = tmp @ P_pred @ tmp.T + K_k @ R_k @ K_k.T

            # Store the results for time step k
            X_kalman[k, :] = x_update
            stds = np.sqrt(np.maximum(np.diag(P_update), min_std))
            X_lower[k, :] = x_update - z * stds
            X_upper[k, :] = x_update + z * stds

            # Move to next step
            x_curr = x_update
            P_curr = P_update

        df_point = pd.DataFrame(X_kalman, index=index, columns=columns)
        df_lower = pd.DataFrame(X_lower, index=index, columns=columns)
        df_upper = pd.DataFrame(X_upper, index=index, columns=columns)

        # Unscale the results if scaling was applied.
        if scale:
            # Multiply each column i by its scale factor.
            scale_series = pd.Series(scales, index=columns)
            df_point = df_point.multiply(scale_series, axis=1)
            df_lower = df_lower.multiply(scale_series, axis=1)
            df_upper = df_upper.multiply(scale_series, axis=1)

        return df_point, df_lower, df_upper

    # -------------
    # MODE 2: PER SERIES (method="per_series")
    # Run T independent 1D Kalman filters—one for each time series.
    # For each forecast step k (looping over S), we update each series (looping over T).
    # -------------
    elif method == "per_series":
        # Prepare output arrays: each is of shape (S, T)
        X_kalman = np.zeros((S, T))
        X_lower = np.zeros((S, T))
        X_upper = np.zeros((S, T))

        # Set up initial states.
        if initial_x is None:
            # Default: each series gets the average of the n models’ forecast at time 0.
            x_0 = np.mean(F[:, 0, :], axis=0)  # shape (T,)
        else:
            x_0 = np.asarray(initial_x, dtype=float)
            if x_0.shape != (T,):
                raise ValueError("initial_x must be shape (T,) for 'per_series' mode")

        if initial_P is None:
            P_0 = np.ones(T)  # one variance per series
        else:
            P_0 = np.array(initial_P, dtype=float)
            if P_0.shape != (T,):
                raise ValueError("initial_P must be shape (T,) for 'per_series' mode")

        # For each series i we store the current state and variance.
        x_curr_all = x_0.copy()
        P_curr_all = P_0.copy()

        # Define the measurement model: since state is scalar, let H be a 1D vector of ones (length n)
        H = np.ones(n)  # now H is shape (n,)

        for k in range(S):
            # For each forecast step, update each series i
            for i in range(T):
                # Time update (prediction)
                x_pred = x_curr_all[i]
                if adapt_Q == "spread" or adapt_R == "spread":
                    var_models = np.var(F[:, k, i], ddof=1)
                else:
                    var_models = 1.0

                base_Q = Q_init if np.isscalar(Q_init) else Q_init[i]
                Q_k = base_Q * var_models if adapt_Q == "spread" else base_Q
                P_pred = P_curr_all[i] + Q_k

                # Measurement update:
                z_k = F[:, k, i]  # n measurements for series i at time k (shape (n,))
                y_k = H * x_pred  # predicted measurement (shape (n,))
                base_R = R_init if np.isscalar(R_init) else R_init[i]
                if adapt_R == "spread":
                    R_k = base_R * var_models * np.eye(n)
                else:
                    R_k = base_R * np.eye(n)
                # Innovation covariance: note use of np.outer to form H*H^T.
                S_k = P_pred * np.outer(H, H) + R_k  # shape (n, n)
                # Kalman gain: K_k is a 1D vector of length n.
                K_k = (P_pred * H) @ robust_pinv(S_k)
                # K_k, residuals, rank, s = np.linalg.lstsq(S_k, P_pred @ H.T, rcond=None)

                # Compute the scalar update; ensure a scalar is returned.
                innovation = z_k - y_k  # shape (n,)
                x_update = x_pred + np.dot(K_k, innovation)
                x_update = float(x_update)  # ensure scalar
                # Update covariance (using a simplified Joseph form)
                tmp = 1 - np.dot(K_k, H)
                P_update = tmp * P_pred + np.dot(K_k, R_k @ K_k)
                P_update = float(P_update)

                # Store and update the state for series i.
                x_curr_all[i] = x_update
                P_curr_all[i] = P_update
                X_kalman[k, i] = x_update
                std_i = np.sqrt(max(P_update, min_std))
                X_lower[k, i] = x_update - z * std_i
                X_upper[k, i] = x_update + z * std_i

        df_point = pd.DataFrame(X_kalman, index=index, columns=columns)
        df_lower = pd.DataFrame(X_lower, index=index, columns=columns)
        df_upper = pd.DataFrame(X_upper, index=index, columns=columns)

        # Unscale the results if scaling was applied.
        if scale:
            scale_series = pd.Series(scales, index=columns)
            df_point = df_point.multiply(scale_series, axis=1)
            df_lower = df_lower.multiply(scale_series, axis=1)
            df_upper = df_upper.multiply(scale_series, axis=1)

        return df_point, df_lower, df_upper

    else:
        raise ValueError(f"Unrecognized method: {method}")



import numpy as np
import pandas as pd


def create_gaussian_wavelet(p, frequency=3, sigma=1.0):
    """
    Create a Gaussian-modulated cosine wavelet with specified frequency and sigma.

    Parameters:
    - p (float): The period or length to generate the wavelet.
    - frequency (int): Frequency of the cosine wave.
    - sigma (float): Standard deviation for the Gaussian envelope.

    Returns:
    - np.ndarray: The generated Gaussian-modulated wavelet.
    """
    x = np.arange(-1, 1, 2 / p)  # Adjusted to accommodate float 'p'
    wavelet = np.cos(frequency * np.pi * x) * np.exp(-(x**2) / (2 * sigma**2))
    return wavelet


def create_morlet_wavelet(p, frequency=3, sigma=1.0):
    """
    Create a Morlet wavelet with specified frequency and sigma.

    Parameters:
    - p (float): The period or length to generate the wavelet.
    - frequency (int): Frequency of the cosine wave.
    - sigma (float): Standard deviation for the Gaussian envelope.

    Returns:
    - np.ndarray: The generated complex Morlet wavelet.
    """
    x = np.arange(-1, 1, 2 / p)  # Adjusted to accommodate float 'p'
    real_part = np.cos(frequency * np.pi * x) * np.exp(-(x**2) / (2 * sigma**2))
    imag_part = np.sin(frequency * np.pi * x) * np.exp(-(x**2) / (2 * sigma**2))
    wavelet = real_part + 1j * imag_part  # Complex wavelet
    return wavelet


def create_real_morlet_wavelet(p, frequency=3, sigma=1.0):
    """
    Create a real-valued Morlet wavelet with specified frequency and sigma.

    Parameters:
    - p (float): The period or length to generate the wavelet.
    - frequency (int): Frequency of the cosine wave.
    - sigma (float): Standard deviation for the Gaussian envelope.

    Returns:
    - np.ndarray: The generated real Morlet wavelet.
    """
    x = np.arange(-1, 1, 2 / p)  # Adjusted to accommodate float 'p'
    # Real component of the Morlet wavelet
    wavelet = np.cos(frequency * np.pi * x) * np.exp(-(x**2) / (2 * sigma**2))
    return wavelet


def create_mexican_hat_wavelet(p, frequency=None, sigma=1.0):
    """
    Create a Mexican Hat wavelet (Ricker wavelet) with specified sigma.

    Parameters:
    - p (float): The period or length to generate the wavelet.
    - sigma (float): Standard deviation for the Gaussian envelope.

    Returns:
    - np.ndarray: The generated Mexican Hat wavelet.
    """
    x = np.arange(-1, 1, 2 / p)  # Adjusted to accommodate float 'p'
    wavelet = (1 - x**2 / sigma**2) * np.exp(-(x**2) / (2 * sigma**2))
    return wavelet


def create_haar_wavelet(p):
    """
    Create a Haar wavelet with specified period `p`.

    Parameters:
    - p (float): The period or length to generate the wavelet.

    Returns:
    - np.ndarray: The generated Haar wavelet.
    """
    if p <= 0:
        raise ValueError("The period `p` must be greater than zero.")

    # Create the Haar wavelet
    x = np.arange(0, p)  # Discrete points to create the wavelet
    # The Haar wavelet has a step function: +1 for the first half, -1 for the second half
    half = len(x) // 2
    wavelet = np.zeros(len(x))
    wavelet[:half] = 1
    wavelet[half:] = -1

    return wavelet


def create_daubechies_db2_wavelet(p):
    """
    Create a Daubechies db2 wavelet with specified period `p`.

    Parameters:
    - p (int): The period or length to generate the wavelet.

    Returns:
    - np.ndarray: The generated Daubechies db2 wavelet.
    """
    if p <= 0:
        raise ValueError("The period `p` must be greater than zero.")

    # Coefficients for the Daubechies db2 wavelet
    # These are the scaling coefficients for the db2 wavelet
    coeffs = np.array(
        [
            (1 + np.sqrt(3)) / 4,
            (3 + np.sqrt(3)) / 4,
            (3 - np.sqrt(3)) / 4,
            (1 - np.sqrt(3)) / 4,
        ]
    )

    # Generate a base wavelet of the specified length `p`
    # To create the wavelet, replicate the coefficients to fit the desired period `p`
    base_wavelet = np.tile(coeffs, int(np.ceil(p / len(coeffs))))[:p]

    return base_wavelet


##############################################################################


def create_wavelet(t, p, sigma=1.0, phase_shift=0, wavelet_type="morlet"):
    """
    Create a real-valued wavelet based on real-world anchored time steps in t,
    with an additional phase shift and a choice of wavelet type.

    Parameters:
    - t (np.ndarray): Array of time steps (in days) from a specified origin.
    - p (float): The period of the wavelet in the same units as t (typically days).
    - sigma (float): Standard deviation for the Gaussian envelope.
    - phase_shift (float): Phase shift to adjust the position of the wavelet peak.
    - wavelet_type (str): Type of wavelet ('morlet' or 'ricker').

    Returns:
    - np.ndarray: The generated wavelet values for each time step.
    """
    x = (t + phase_shift) % p - p / 2  # Normalize and center t around 0

    if wavelet_type == "morlet":
        return np.cos(2 * np.pi * x / p) * np.exp(-(x**2) / (2 * sigma**2))
    elif wavelet_type == "ricker":
        # Ricker (Mexican Hat) wavelet calculation
        a = 2 * sigma**2
        return (1 - (x**2 / a)) * np.exp(-(x**2) / (2 * sigma**2))
    else:
        raise ValueError("Unsupported wavelet type. Choose 'morlet' or 'ricker'.")


def offset_wavelet(p, t, order=5, sigma=1.0, wavelet_type="morlet"):
    """
    Create an offset collection of wavelets with `order` offsets, ensuring that
    peaks are spaced p/order apart.

    Parameters:
    - p (float): Period of the wavelet in the same units as t (typically days).
    - t (np.ndarray): Array of time steps.
    - order (int): The number of offsets.
    - sigma (float): Standard deviation for the Gaussian envelope.
    - wavelet_type (str): Type of wavelet ('morlet' or 'ricker').

    Returns:
    - np.ndarray: A 2D array with `order` wavelets along axis 1.
    """
    wavelet_features = []
    phase_offsets = np.linspace(
        0, p, order, endpoint=False
    )  # Properly space phase shifts over one period

    for phase_shift in phase_offsets:
        wavelet = create_wavelet(t, p, sigma, phase_shift, wavelet_type)
        wavelet_features.append(wavelet)

    return np.stack(wavelet_features, axis=1)


if False:
    DTindex = pd.date_range("2020-01-01", "2024-01-01", freq="D")
    origin_ts = "2030-01-01"
    t = (DTindex - pd.Timestamp(origin_ts)).total_seconds() / 86400

    p = 7
    weekly_wavelets = offset_wavelet(
        p=p,  # Weekly period
        t=t,  # A full year (365 days)
        # origin_ts=origin_ts,
        order=7,  # One offset for each day of the week
        # frequency=2 * np.pi / p,  # Frequency for weekly pattern
        sigma=0.5,  # Smaller sigma for tighter weekly spread
        wavelet_type="morlet",
    )

    # Example for yearly seasonality
    p = 365.25
    yearly_wavelets = offset_wavelet(
        p=p,  # Yearly period
        t=t,  # Three full years
        # origin_ts=origin_ts,
        order=12,  # One offset for each month
        # frequency=2 * np.pi / p,  # Frequency for yearly pattern
        sigma=2.0,  # Larger sigma for broader yearly spread
        wavelet_type="morlet",
    )
    yearly_wavelets2 = offset_wavelet(
        p=p,  # Yearly period
        t=t[-100:],  # Three full years
        # origin_ts=origin_ts,
        order=12,  # One offset for each month
        # frequency=2 * np.pi / p,  # Frequency for yearly pattern
        sigma=2.0,  # Larger sigma for broader yearly spread
        wavelet_type="morlet",
    )
    print(np.allclose(yearly_wavelets[-100:], yearly_wavelets2))

    # Display wavelet patterns for visualization
    import matplotlib.pyplot as plt

    plt.figure(figsize=(12, 6))
    pd.DataFrame(weekly_wavelets).plot(title="Weekly Wavelets", ax=plt.gca())
    pd.DataFrame(yearly_wavelets).plot(title="Yearly Wavelets", ax=plt.gca())
    plt.show()

    pd.DataFrame(weekly_wavelets[0:50]).plot(title="Weekly Wavelets", ax=plt.gca())
    plt.show()


##############################################################################


def continuous_db2_wavelet(t, p, order, sigma):
    # Normalize t to [0, 1) interval based on period p, scaled by order to include multiple cycles
    x = (order * t % p) / p
    if order % 3 == 0:
        x = x + 0.3
    gaussian_envelope = np.exp(-0.5 * ((x - 0.5) / sigma) ** 2)
    sinusoidal_component = np.sin(2 * np.pi * x)
    wavelet = gaussian_envelope * sinusoidal_component
    return wavelet


def create_narrowing_wavelets(p, max_order, t, sigma=0.5):
    wavelets = []
    for order in range(1, max_order + 1):
        sigma = sigma / order  # Narrow the Gaussian envelope as order increases
        wavelet = continuous_db2_wavelet(t, p, order, sigma)
        wavelets.append(wavelet)
    return np.array(wavelets).T


if False:
    # Example usage
    DTindex = pd.date_range("2020-01-01", "2024-01-01", freq="D")
    origin_ts = "2020-01-01"
    t_full = (DTindex - pd.Timestamp(origin_ts)).total_seconds() / 86400

    p = 365.25  # Example period
    max_order = 5  # Example maximum order

    # Full set of wavelets
    wavelets = create_narrowing_wavelets(p, max_order, t_full)

    # Wavelets for the last 100 days
    t_subset = t_full[-100:]
    wavelet_short = create_narrowing_wavelets(p, max_order, t_subset)

    # Check if the last 100 days of the full series match the subset
    print(np.allclose(wavelets[-100:], wavelet_short))  # This should be true

    # Plotting the wavelets
    plt.figure(figsize=(12, 6))
    for i in range(max_order):
        plt.plot(DTindex[-100:], wavelets[-100:, i], label=f"Order {i+1}")
        plt.plot(
            DTindex[-100:],
            wavelet_short[:, i],
            label=f"Subset Order {i+1}",
            linestyle="--",
        )
    plt.title("Comparison of Full Wavelets and Subset")
    plt.legend()
    plt.show()



# -*- coding: utf-8 -*-
"""
Seasonal unittests.

Created on Sat May  4 21:43:02 2024

@author: Colin
"""

import unittest
import numpy as np
import pandas as pd
from autots.tools.seasonal import date_part, base_seasonalities, datepart_components, random_datepart, fourier_df
from autots.tools.holiday import holiday_flag
from autots.tools.wavelet import create_narrowing_wavelets, offset_wavelet


class TestSeasonal(unittest.TestCase):

    def test_date_part(self):
        DTindex = pd.date_range("2020-01-01", "2024-01-01", freq="D")
        for method in base_seasonalities:
            df = date_part(DTindex, method=method, set_index=True)
            self.assertEqual(df.shape[0], DTindex.shape[0])
            self.assertGreater(df.shape[1], 1)

    def test_date_components(self):
        DTindex = pd.date_range("2023-01-01", "2024-01-01", freq="h")
        for method in datepart_components:
            df = date_part(DTindex, method=method, set_index=True)
            self.assertEqual(df.shape[0], DTindex.shape[0])

    def test_random_datepart(self):
        out = random_datepart()
        self.assertTrue(out)

    def test_fourier(self):
        DTindex = pd.date_range("2020-01-02", "2024-01-01", freq="D")
        order = 10
        df = fourier_df(DTindex, seasonality=365.25, order=order)
        self.assertEqual(df.shape[1], order * 2)
        self.assertEqual(df.shape[0], DTindex.shape[0])
        self.assertAlmostEqual(df.mean().sum(), 0.0)

    def test_wavelets_repeat(self):
        DTindex = pd.date_range("2020-01-01", "2024-01-01", freq="D")
        origin_ts = "2030-01-01"
        t = (DTindex - pd.Timestamp(origin_ts)).total_seconds() / 86400

        p = 7
        w_order = 7
        weekly_wavelets = offset_wavelet(
            p=p,  # Weekly period
            t=t,  # A full year (365 days)
            # origin_ts=origin_ts,
            order=w_order,  # One offset for each day of the week
            # frequency=2 * np.pi / p,  # Frequency for weekly pattern
            sigma=0.5,  # Smaller sigma for tighter weekly spread
            wavelet_type="ricker",
        )
        self.assertEqual(weekly_wavelets.shape[1], w_order)

        # Example for yearly seasonality
        p = 365.25
        y_order = 12
        yearly_wavelets = offset_wavelet(
            p=p,  # Yearly period
            t=t,  # Three full years
            # origin_ts=origin_ts,
            order=y_order,  # One offset for each month
            # frequency=2 * np.pi / p,  # Frequency for yearly pattern
            sigma=2.0,  # Larger sigma for broader yearly spread
            wavelet_type="morlet",
        )
        yearly_wavelets2 = offset_wavelet(
            p=p,  # Yearly period
            t=t[-100:],  # Three full years
            # origin_ts=origin_ts,
            order=y_order,  # One offset for each month
            # frequency=2 * np.pi / p,  # Frequency for yearly pattern
            sigma=2.0,  # Larger sigma for broader yearly spread
            wavelet_type="morlet",
        )
        self.assertEqual(yearly_wavelets.shape[1], y_order)
        self.assertTrue(np.allclose(yearly_wavelets[-100:], yearly_wavelets2))

    def test_wavelet_continuous(self):
        DTindex = pd.date_range("2020-01-01", "2024-01-01", freq="D")
        origin_ts = "2020-01-01"
        t_full = (DTindex - pd.Timestamp(origin_ts)).total_seconds() / 86400

        p = 365.25  # Example period
        max_order = 5  # Example maximum order

        # Full set of wavelets
        wavelets = create_narrowing_wavelets(p, max_order, t_full)

        # Wavelets for the last 100 days
        t_subset = t_full[-100:]
        wavelet_short = create_narrowing_wavelets(p, max_order, t_subset)

        # Check if the last 100 days of the full series match the subset
        self.assertTrue(np.allclose(wavelets[-100:], wavelet_short))

    def test_holiday_flag(self):
        # hourly being trickier
        train_index = pd.date_range("2020-01-01", "2023-01-01", freq="h")
        pred_index = pd.date_range("2023-01-02", "2024-01-01", freq="h")
        
        train_holiday = holiday_flag(train_index, country=["US", "CA"], encode_holiday_type=True)
        pred_holiday = holiday_flag(pred_index, country=["US", "CA"], encode_holiday_type=True)

        self.assertCountEqual(train_holiday.columns.tolist(), pred_holiday.columns.tolist())
        self.assertGreaterEqual(train_holiday.sum().sum(), 24)
        self.assertGreaterEqual(pred_holiday.sum().sum(), 24)



# -*- coding: utf-8 -*-
"""
Base model information

@author: Colin
"""
import json
import random
import warnings
import datetime
import numpy as np
import pandas as pd
from autots.tools.constraint import apply_constraint_single
from autots.tools.shaping import infer_frequency, clean_weights
from autots.evaluator.metrics import full_metric_evaluation


def create_forecast_index(frequency, forecast_length, train_last_date, last_date=None):
    if frequency == 'infer':
        raise ValueError(
            "create_forecast_index run without specific frequency, run basic_profile first or pass proper frequency to model init"
        )
    return pd.date_range(
        freq=frequency,
        start=train_last_date if last_date is None else last_date,
        periods=int(forecast_length + 1),
    )[
        1:
    ]  # note the disposal of the first (already extant) date


class ModelObject(object):
    """Generic class for holding forecasting models.

    Models should all have methods:
        .fit(df, future_regressor = []) (taking a DataFrame with DatetimeIndex and n columns of n timeseries)
        .predict(forecast_length = int, future_regressor = [], just_point_forecast = False)
        .get_new_params() - return a dictionary of weighted random selected parameters

    Args:
        name (str): Model Name
        frequency (str): String alias of datetime index frequency or else 'infer'
        prediction_interval (float): Confidence interval for probabilistic forecast
        n_jobs (int): used by some models that parallelize to multiple cores
    """

    def __init__(
        self,
        name: str = "Uninitiated Model Name",
        frequency: str = 'infer',
        prediction_interval: float = 0.9,
        regression_type: str = None,
        fit_runtime=datetime.timedelta(0),
        holiday_country: str = 'US',
        random_seed: int = 2020,
        verbose: int = 0,
        n_jobs: int = -1,
    ):
        self.name = name
        self.frequency = frequency
        self.prediction_interval = prediction_interval
        self.regression_type = regression_type
        self.fit_runtime = fit_runtime
        self.holiday_country = holiday_country
        self.random_seed = random_seed
        self.verbose = verbose
        self.verbose_bool = True if self.verbose > 1 else False
        self.n_jobs = n_jobs

    def __repr__(self):
        """Print."""
        return 'ModelObject of ' + self.name + ' uses standard .fit/.predict'

    def basic_profile(self, df):
        """Capture basic training details."""
        if 0 in df.shape:
            raise ValueError(f"{self.name} training dataframe has no data: {df.shape}")
        self.startTime = datetime.datetime.now()
        self.train_shape = df.shape
        self.column_names = df.columns
        self.train_last_date = df.index[-1]
        if self.frequency == 'infer':
            self.frequency = infer_frequency(df.index)

        return df

    def create_forecast_index(self, forecast_length: int, last_date=None):
        """Generate a pd.DatetimeIndex appropriate for a new forecast.

        Warnings:
            Requires ModelObject.basic_profile() being called as part of .fit()
        """

        return create_forecast_index(
            self.frequency, forecast_length, self.train_last_date, last_date
        )

    def get_params(self):
        """Return dict of current parameters."""
        return {}

    def get_new_params(self, method: str = 'random'):
        """Return dict of new parameters for parameter tuning."""
        return {}

    def fit_data(self, df, future_regressor=None):
        self.basic_profile(df)
        if future_regressor is not None:
            self.regressor_train = future_regressor
        return self

    @staticmethod
    def time():
        return datetime.datetime.now()


def apply_constraints(
    forecast,
    lower_forecast,
    upper_forecast,
    constraints=None,
    df_train=None,
    # old args
    constraint_method=None,
    constraint_regularization=None,
    upper_constraint=None,
    lower_constraint=None,
    bounds=True,
):
    """Use constraint thresholds to adjust outputs by limit.

    Args:
        forecast (pd.DataFrame): forecast df, wide style
        lower_forecast (pd.DataFrame): lower bound forecast df
            if bounds is False, upper and lower forecast dataframes are unused and can be empty
        upper_forecast (pd.DataFrame): upper bound forecast df
        constraints (list): list of dictionaries of constraints to apply
            keys: "constraint_method" (same as below, old args), "constraint_regularization", "constraint_value", "constraint_direction" (upper/lower), bounds
        df_train (pd.DataFrame): required for quantile/stdev methods to find threshold values
        # old args
        constraint_method (str): one of
            stdev_min - threshold is min and max of historic data +/- constraint * st dev of data
            stdev - threshold is the mean of historic data +/- constraint * st dev of data
            absolute - input is array of length series containing the threshold's final value for each
            quantile - constraint is the quantile of historic data to use as threshold
            last_window - certain percentage above and below the last n data values
            slope - cannot exceed a certain growth rate from last historical value
        constraint_regularization (float): 0 to 1
            where 0 means no constraint, 1 is hard threshold cutoff, and in between is penalty term
        upper_constraint (float): or array, depending on method, None if unused
        lower_constraint (float): or array, depending on method, None if unused
        bounds (bool): if True, apply to upper/lower forecast, otherwise False applies only to forecast

    Returns:
        forecast, lower, upper (pd.DataFrame)
    """
    # handle old style
    if constraint_method is not None:
        if constraints is not None:
            raise ValueError(
                f"both constraint_method (old way) and constraints (new way) args passed, this will not work. Constraints: {constraints}"
            )
        else:
            constraints = []
            if upper_constraint is not None:
                constraints.append(
                    {
                        "constraint_method": constraint_method,
                        "constraint_value": upper_constraint,
                        "constraint_direction": "upper",
                        "constraint_regularization": constraint_regularization,
                        "bounds": bounds,
                    }
                )
            if lower_constraint is not None:
                constraints.append(
                    {
                        "constraint_method": constraint_method,
                        "constraint_value": lower_constraint,
                        "constraint_direction": "lower",
                        "constraint_regularization": constraint_regularization,
                        "bounds": bounds,
                    }
                )
        print(constraints)
    if constraints is None or not constraints:
        print("no constraint applied")
        return forecast, lower_forecast, upper_forecast
    if isinstance(constraints, dict):
        constraints = [constraints]
    for constraint in constraints:
        forecast, lower_forecast, upper_forecast = apply_constraint_single(
            forecast, lower_forecast, upper_forecast, df_train=df_train, **constraint
        )

    return forecast, lower_forecast, upper_forecast


def extract_single_series_from_horz(series, model_name, model_parameters):
    title_prelim = str(model_name)[0:80]
    if title_prelim == "Ensemble":
        ensemble_type = model_parameters.get('model_name', "Ensemble")
        # horizontal and mosaic ensembles
        if "series" in model_parameters.keys():
            model_id = model_parameters['series'].get(series, "Horizontal")
            if isinstance(model_id, dict):
                model_id = list(model_id.values())
            if not isinstance(model_id, list):
                model_id = [str(model_id)]
            res = []
            for imod in model_id:
                res.append(
                    model_parameters.get("models", {})
                    .get(imod, {})
                    .get('Model', "Horizontal")
                )
            title_prelim = ", ".join(set(res))
            if len(model_id) > 1:
                title_prelim = "Mosaic: " + str(title_prelim)
        else:
            title_prelim = ensemble_type
    return str(title_prelim)


def extract_single_transformer(
    series, model_name, model_parameters, transformation_params
):
    if isinstance(transformation_params, str):
        transformation_params = json.loads(transformation_params)
    if isinstance(model_parameters, str):
        model_parameters = json.loads(model_parameters)
    if model_name == "Ensemble":
        # horizontal and mosaic ensembles
        if "series" in model_parameters.keys():
            model_id = model_parameters['series'].get(series, "Horizontal")
            if isinstance(model_id, dict):
                model_id = list(model_id.values())
            if not isinstance(model_id, list):
                model_id = [str(model_id)]
            res = []
            for imod in model_id:
                chosen_mod = model_parameters.get("models", {}).get(imod, {})
                res.append(
                    extract_single_transformer(
                        series,
                        chosen_mod.get("Model"),
                        chosen_mod.get("ModelParameters"),
                        transformation_params=chosen_mod.get(
                            "TransformationParameters"
                        ),
                    )
                )
            return ", ".join(res)
        allz = []
        for idz, mod in model_parameters.get("models").items():
            allz.append(
                extract_single_transformer(
                    series,
                    mod.get("Model"),
                    mod.get("ModelParameters"),
                    transformation_params=mod.get("TransformationParameters"),
                )
            )
        return ", ".join(allz)
    else:
        if isinstance(transformation_params, dict):
            trans_dict = transformation_params.get("transformations")
            if isinstance(trans_dict, dict):
                return ", ".join(list(trans_dict.values()))
            else:
                return "None"
        else:
            return "None"


def create_seaborn_palette_from_cmap(cmap_name="gist_rainbow", n=10):
    import matplotlib.pyplot as plt
    import seaborn as sns

    # Get the colormap from matplotlib
    cm = plt.get_cmap(cmap_name)

    # Create a range of colors from the colormap
    colors = cm(np.linspace(0, 1, n))

    # Convert to a seaborn palette
    palette = sns.color_palette(colors)

    return palette


# Function to calculate the peak density of each model's distribution
def calculate_peak_density(model, data, group_col='Model', y_col='TotalRuntimeSeconds'):
    from scipy.stats import gaussian_kde

    model_data = data[data[group_col] == model][y_col]
    kde = gaussian_kde(model_data)
    return np.max(kde(model_data))


def plot_distributions(
    runtimes_data,
    group_col='Model',
    y_col='TotalRuntimeSeconds',
    xlim=None,
    xlim_right=None,
    title_suffix="",
):
    import matplotlib.pyplot as plt
    import seaborn as sns

    single_obs_models = runtimes_data.groupby(group_col).filter(lambda x: len(x) == 1)
    multi_obs_models = runtimes_data.groupby(group_col).filter(lambda x: len(x) > 1)

    # Calculate the average peak density across all models with multiple observations
    average_peak_density = np.mean(
        [
            calculate_peak_density(model, multi_obs_models, group_col, y_col)
            for model in multi_obs_models[group_col].unique()
        ]
    )

    # Correcting the color palette to match the number of unique models
    unique_models = runtimes_data[group_col].nunique()
    # palette = sns.color_palette("tab10", n_colors=unique_models)
    palette = create_seaborn_palette_from_cmap("gist_rainbow", n=unique_models)
    sorted_models = runtimes_data[group_col].value_counts().index.tolist()
    # sorted_models = runtimes_data[group_col].unique()
    zip_palette = dict(zip(sorted_models, palette))

    # Create a new figure for the plot
    fig = plt.figure(figsize=(12, 8))

    # Plot the density plots for multi-observation models
    density_plot = sns.kdeplot(  # noqa
        data=multi_obs_models,
        x=y_col,
        hue=group_col,
        fill=True,
        common_norm=False,
        palette=zip_palette,
        alpha=0.5,
    )

    # Plot the points for single-observation models at the average peak density
    if not single_obs_models.empty:
        point_plot = sns.scatterplot(  # noqa
            data=single_obs_models,
            x=y_col,
            y=[average_peak_density] * len(single_obs_models),
            hue=group_col,
            palette=zip_palette,
            legend=False,
            marker='o',  # s=10
        )
    # Adjusting legend - Manually combining elements
    handles, labels = [], []
    for model, color in zip_palette.items():
        handles.append(plt.Line2D([0], [0], linestyle="none", c=color, marker='o'))
        labels.append(model)

    # Create the combined legend
    plt.legend(handles, labels, title=group_col)  # , bbox_to_anchor=(1.05, 1), loc=2

    # Adding titles and labels
    plt.title(f'Distribution of {y_col} by {group_col}{title_suffix}', fontsize=16)
    plt.xlabel(f'{y_col}', fontsize=14)
    plt.ylabel('Density', fontsize=14)

    # Adjust layout
    plt.tight_layout()
    if xlim is not None:
        plt.xlim(left=xlim)
    if xlim_right is not None:
        plt.xlim(right=runtimes_data[y_col].quantile(xlim_right))

    return fig


class PredictionObject(object):
    """Generic class for holding forecast information.

    Attributes:
        model_name
        model_parameters
        transformation_parameters
        forecast
        upper_forecast
        lower_forecast

    Methods:
        long_form_results: return complete results in long form
        total_runtime: return runtime for all model components in seconds
        plot
        evaluate
        apply_constraints
    """

    def __init__(
        self,
        model_name: str = 'Uninitiated',
        forecast_length: int = 0,
        forecast_index=np.nan,
        forecast_columns=np.nan,
        lower_forecast=np.nan,
        forecast=np.nan,
        upper_forecast=np.nan,
        prediction_interval: float = 0.9,
        predict_runtime=datetime.timedelta(0),
        fit_runtime=datetime.timedelta(0),
        model_parameters={},
        transformation_parameters={},
        transformation_runtime=datetime.timedelta(0),
        per_series_metrics=np.nan,
        per_timestamp=np.nan,
        avg_metrics=np.nan,
        avg_metrics_weighted=np.nan,
        full_mae_error=None,
        model=None,
        transformer=None,
    ):
        self.model_name = self.name = model_name
        self.model_parameters = model_parameters
        self.transformation_parameters = transformation_parameters
        self.forecast_length = forecast_length
        self.forecast_index = forecast_index
        self.forecast_columns = forecast_columns
        self.lower_forecast = lower_forecast
        self.forecast = forecast
        self.upper_forecast = upper_forecast
        self.prediction_interval = prediction_interval
        self.predict_runtime = predict_runtime
        self.fit_runtime = fit_runtime
        self.transformation_runtime = transformation_runtime
        # eval attributes
        self.per_series_metrics = per_series_metrics
        self.per_timestamp = per_timestamp
        self.avg_metrics = avg_metrics
        self.avg_metrics_weighted = avg_metrics_weighted
        self.full_mae_error = full_mae_error
        # model attributes, not normally used
        self.model = model
        self.transformer = transformer
        self.runtime_dict = None

    def __repr__(self):
        """Print."""
        if isinstance(self.forecast, pd.DataFrame):
            return "Prediction object: \nReturn .forecast, \n .upper_forecast, \n .lower_forecast \n .model_parameters \n .transformation_parameters"
        else:
            return "Empty prediction object."

    def __bool__(self):
        """bool version of class."""
        if isinstance(self.forecast, pd.DataFrame):
            return True
        else:
            return False

    def long_form_results(
        self,
        id_name="SeriesID",
        value_name="Value",
        interval_name='PredictionInterval',
        update_datetime_name=None,
        datetime_column=None,
    ):
        """Export forecasts (including upper and lower) as single 'long' format output

        Args:
            id_name (str): name of column containing ids
            value_name (str): name of column containing numeric values
            interval_name (str): name of column telling you what is upper/lower
            datetime_column (str): if None, is index, otherwise, name of column for datetime
            update_datetime_name (str): if not None, adds column with current timestamp and this name

        Returns:
            pd.DataFrame
        """
        upload = pd.melt(
            self.forecast.rename_axis(index='datetime').reset_index(),
            var_name=id_name,
            value_name=value_name,
            id_vars="datetime",
        ).set_index("datetime")
        upload[interval_name] = "50%"
        upload_upper = pd.melt(
            self.upper_forecast.rename_axis(index='datetime').reset_index(),
            var_name=id_name,
            value_name=value_name,
            id_vars="datetime",
        ).set_index("datetime")
        upload_upper[interval_name] = (
            f"{round(100 - ((1- self.prediction_interval)/2) * 100, 0)}%"
        )
        upload_lower = pd.melt(
            self.lower_forecast.rename_axis(index='datetime').reset_index(),
            var_name=id_name,
            value_name=value_name,
            id_vars="datetime",
        ).set_index("datetime")
        upload_lower[interval_name] = (
            f"{round(((1- self.prediction_interval)/2) * 100, 0)}%"
        )

        upload = pd.concat([upload, upload_upper, upload_lower], axis=0)
        if datetime_column is not None:
            upload.index.name = str(datetime_column)
            upload = upload.reset_index(drop=False)
        if update_datetime_name is not None:
            upload[update_datetime_name] = datetime.datetime.utcnow()
        return upload

    def total_runtime(self):
        """Combine runtimes."""
        return self.fit_runtime + self.predict_runtime + self.transformation_runtime

    def extract_ensemble_runtimes(self):
        """Return a dataframe of final runtimes per model for standard ensembles."""
        if self.runtime_dict is None or not bool(self.model_parameters):
            return None
        else:
            runtimes = pd.DataFrame(
                self.runtime_dict.items(), columns=['ID', 'Runtime']
            )
            runtimes['TotalRuntimeSeconds'] = runtimes['Runtime'].dt.total_seconds()
            new_models = {
                x: y.get("Model")
                for x, y in self.model_parameters.get("models").items()
            }
            models = pd.DataFrame(new_models.items(), columns=['ID', 'Model'])
            return runtimes.merge(models, how='left', on='ID')

    def plot_ensemble_runtimes(self, xlim_right=None):
        """Plot ensemble runtimes by model type."""
        runtimes_data = self.extract_ensemble_runtimes()

        if runtimes_data is None:
            return None
        else:
            return plot_distributions(
                runtimes_data,
                group_col='Model',
                y_col='TotalRuntimeSeconds',
                xlim=0,
                xlim_right=xlim_right,
                title_suffix=" in Chosen Ensemble",
            )

    def plot_df(
        self,
        df_wide=None,
        series: str = None,
        remove_zeroes: bool = False,
        interpolate: str = None,
        start_date: str = None,
    ):
        if series is None:
            series = random.choice(self.forecast.columns)

        model_name = self.model_name
        if model_name == "Ensemble":
            if 'series' in self.model_parameters.keys():
                if "profile" in self.model_parameters["model_metric"]:
                    from autots.tools.profile import profile_time_series

                    df = df_wide if df_wide is not None else self.forecast
                    profile = profile_time_series(df)
                    # I'm not sure why I made it sometimes coming as ID and sometimes SERIES...
                    if "ID" in profile.columns:
                        key_col = "ID"
                    else:
                        key_col = "SERIES"
                    h_params = self.model_parameters['series'][
                        profile[profile[key_col] == series]["PROFILE"].iloc[0]
                    ]
                else:
                    h_params = self.model_parameters['series'][series]
                if isinstance(h_params, str):
                    model_name = self.model_parameters['models'][h_params]['Model']

        if df_wide is not None:
            plot_df = pd.DataFrame(
                {
                    'actuals': df_wide[series],
                    'up_forecast': self.upper_forecast[series],
                    'low_forecast': self.lower_forecast[series],
                    'forecast': self.forecast[series],
                }
            )
        else:
            plot_df = pd.DataFrame(
                {
                    'up_forecast': self.upper_forecast[series],
                    'low_forecast': self.lower_forecast[series],
                    'forecast': self.forecast[series],
                }
            )
        if remove_zeroes:
            plot_df[plot_df == 0] = np.nan
        if interpolate is not None:
            plot_df["actuals"] = plot_df["actuals"].interpolate(
                method=interpolate, limit_direction="backward"
            )
            plot_df["forecast"] = plot_df["forecast"].interpolate(
                method=interpolate, limit_direction="backward", limit=5
            )

        if start_date is not None:
            start_date = pd.to_datetime(start_date)
            if plot_df.index.max() < start_date:
                raise ValueError("start_date is more recent than all data provided")
            plot_df = plot_df[plot_df.index >= start_date]
        return plot_df

    def plot(
        self,
        df_wide=None,
        series: str = None,
        remove_zeroes: bool = False,
        interpolate: str = None,
        start_date: str = "auto",
        alpha=0.3,
        facecolor="black",
        loc="upper right",
        title=None,
        title_substring=None,
        vline=None,
        colors=None,
        include_bounds=True,
        **kwargs,
    ):
        """Generate an example plot of one series. Does not handle non-numeric forecasts.

        Args:
            df_wide (str): historic data for plotting actuals
            series (str): column name of series to plot. Random if None.
            ax: matplotlib axes to pass through to pd.plot()
            remove_zeroes (bool): if True, don't plot any zeroes
            interpolate (str): if not None, a method to pass to pandas interpolate
            start_date (str): Y-m-d string or Timestamp to remove all data before
            vline (datetime): datetime of dashed vertical line to plot
            colors (dict): colors mapping dictionary col: color
            alpha (float): intensity of bound interval shading
            title (str): title
            title_substring (str): additional title details to pass to existing, moves series name to axis
            include_bounds (bool): if True, shows region of upper and lower forecasts
            **kwargs passed to pd.DataFrame.plot()
        """
        if start_date == "auto":
            if df_wide is not None:
                slx = -self.forecast_length * 3
                if abs(slx) > df_wide.shape[0]:
                    slx = 0
                start_date = df_wide.index[slx]
            else:
                start_date = self.forecast.index[0]

        if series is None:
            series = random.choice(self.forecast.columns)
        plot_df = self.plot_df(
            df_wide=df_wide,
            series=series,
            remove_zeroes=remove_zeroes,
            interpolate=interpolate,
            start_date=start_date,
        )
        if self.forecast_length == 1 and 'actuals' in plot_df.columns:
            if plot_df.shape[0] > 3:
                plot_df['forecast'].iloc[-2] = plot_df['actuals'].iloc[-2]
        if 'actuals' not in plot_df.columns:
            plot_df['actuals'] = np.nan
        if colors is None:
            colors = {
                'low_forecast': '#A5ADAF',
                'up_forecast': '#A5ADAF',
                'forecast': '#003399',  # '#4D4DFF',
                'actuals': '#AFDBF5',
            }
        if title is None:
            title_prelim = extract_single_series_from_horz(
                series,
                model_name=self.model_name,
                model_parameters=self.model_parameters,
            )[0:80]
            if title_substring is None:
                title = f"{series} with model {title_prelim}"
            else:
                title = f"{title_substring} with model {title_prelim}"

        ax = plot_df[['actuals', 'forecast']].plot(title=title, color=colors, **kwargs)
        if include_bounds:
            ax.fill_between(
                plot_df.index,
                plot_df['up_forecast'],
                plot_df['low_forecast'],
                alpha=alpha,
                color="#A5ADAF",
                label="Prediction Interval",
            )
        if vline is not None:
            ax.vlines(
                x=vline,
                ls='--',
                lw=1,
                colors='darkred',
                ymin=plot_df.min().min(),
                ymax=plot_df.max().max(),
            )
            # ax.text(vline, plot_df.max().max(), "Event", color='darkred', verticalalignment='top')
        if title_substring is not None:
            ax.set_ylabel(series)
        # ax.grid(True, which="both", ls="--", linewidth=0.5)
        # ax.legend(loc=loc)
        return ax

    def plot_grid(
        self,
        df_wide=None,
        start_date='auto',
        interpolate=None,
        remove_zeroes=False,
        figsize=(24, 18),
        title="AutoTS Forecasts",
        cols=None,
        series=None,  # alias for above
        colors=None,
        include_bounds=True,
    ):
        """Plots multiple series in a grid, if present. Mostly identical args to the single plot function."""
        import matplotlib.pyplot as plt

        if series is not None and cols is None:
            cols = series
        if cols is None:
            cols = self.forecast.columns.tolist()
        num_cols = len(cols)
        if num_cols > 4:
            nrow = 2
            ncol = 3
            num_cols = 6
        elif num_cols > 2:
            nrow = 2
            ncol = 2
        else:
            nrow = 1
            ncol = 2
        fig, axes = plt.subplots(nrow, ncol, figsize=figsize, constrained_layout=True)
        fig.suptitle(title, fontsize='xx-large')
        count = 0
        if len(cols) != num_cols:
            sample_cols = random.choices(cols, k=num_cols)
        else:
            sample_cols = cols
        for r in range(nrow):
            for c in range(ncol):
                if nrow > 1:
                    ax = axes[r, c]
                else:
                    ax = axes[c]
                if count + 1 > num_cols:
                    pass
                else:
                    col = sample_cols[count]
                    self.plot(
                        df_wide=df_wide,
                        series=col,
                        remove_zeroes=remove_zeroes,
                        interpolate=interpolate,
                        start_date=start_date,
                        colors=colors,
                        include_bounds=include_bounds,
                        ax=ax,
                    )
                    count += 1
        return fig

    def evaluate(
        self,
        actual,
        series_weights: dict = None,
        df_train=None,
        per_timestamp_errors: bool = False,
        full_mae_error: bool = True,
        scaler=None,
        cumsum_A=None,
        diff_A=None,
        last_of_array=None,
        column_names=None,
        custom_metric=None,
    ):
        """Evalute prediction against test actual. Fills out attributes of base object.

        This fails with pd.NA values supplied.

        Args:
            actual (pd.DataFrame): dataframe of actual values of (forecast length * n series)
            series_weights (dict): key = column/series_id, value = weight
            df_train (pd.DataFrame): historical values of series, wide,
                used for setting scaler for SPL
                necessary for MADE and Contour if forecast_length == 1
                if None, actuals are used instead (suboptimal).
            per_timestamp (bool): whether to calculate and return per timestamp direction errors
            custom metric (callable): a function to generate a custom metric. Expects func(A, F, df_train, prediction_interval) where the first three are np arrays of wide style 2d.

        Returns:
            per_series_metrics (pandas.DataFrame): contains a column for each series containing accuracy metrics
            per_timestamp (pandas.DataFrame): smape accuracy for each timestamp, avg of all series
            avg_metrics (pandas.Series): average values of accuracy across all input series
            avg_metrics_weighted (pandas.Series): average values of accuracy across all input series weighted by series_weight, if given
            full_mae_errors (numpy.array): abs(actual - forecast)
            scaler (numpy.array): precomputed scaler for efficiency, avg value of series in order of columns
        """
        # some forecasts have incorrect columns (they shouldn't but they do as a bug sometimes)
        if column_names is not None:
            use_cols = column_names
        elif isinstance(df_train, pd.DataFrame):
            use_cols = df_train.columns
        else:
            use_cols = self.forecast_columns
        # arrays are faster for math than pandas dataframes
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            (
                self.per_series_metrics,
                self.full_mae_errors,
                self.squared_errors,
                self.upper_pl,
                self.lower_pl,
            ) = full_metric_evaluation(
                A=actual,
                F=self.forecast,
                upper_forecast=self.upper_forecast,
                lower_forecast=self.lower_forecast,
                df_train=df_train,
                prediction_interval=self.prediction_interval,
                columns=use_cols,
                scaler=scaler,
                return_components=True,
                cumsum_A=cumsum_A,
                diff_A=diff_A,
                last_of_array=last_of_array,
                custom_metric=custom_metric,
            )

        if per_timestamp_errors:
            smape_df = abs(self.forecast - actual) / (abs(self.forecast) + abs(actual))
            weight_mean = np.mean(list(series_weights.values()))
            wsmape_df = (smape_df * series_weights) / weight_mean
            smape_cons = (np.nansum(wsmape_df, axis=1) * 200) / np.count_nonzero(
                ~np.isnan(actual), axis=1
            )
            per_timestamp = pd.DataFrame({'weighted_smape': smape_cons}).transpose()
            self.per_timestamp = per_timestamp

        # check series_weights information
        if series_weights is None:
            series_weights = clean_weights(weights=False, series=self.forecast.columns)
        # make sure the series_weights are passed correctly to metrics
        if len(series_weights) != self.forecast.shape[1]:
            series_weights = {col: series_weights[col] for col in self.forecast.columns}

        # this weighting won't work well if entire metrics are NaN
        # but results should still be comparable
        self.avg_metrics_weighted = (self.per_series_metrics * series_weights).sum(
            axis=1, skipna=True
        ) / sum(series_weights.values())
        self.avg_metrics = self.per_series_metrics.mean(axis=1, skipna=True)
        if False:
            # vn1 temporary
            submission = self.forecast
            objective = actual
            abs_err = np.nansum(np.abs(submission - objective))
            err = np.nansum((submission - objective))
            score = abs_err + abs(err)
            epsilon = 1
            big_sum = (
                np.nan_to_num(objective, nan=0.0, posinf=0.0, neginf=0.0).sum().sum()
                + epsilon
            )
            score /= big_sum
            self.avg_metrics["custom"] = score
            self.avg_metrics_weighted["custom"] = score
        return self

    def apply_constraints(
        self,
        constraints=None,
        df_train=None,
        # old args
        constraint_method=None,
        constraint_regularization=None,
        upper_constraint=None,
        lower_constraint=None,
        bounds=True,
    ):
        """Use constraint thresholds to adjust outputs by limit.

        Example:
            apply_constraints(
                constraints=[
                    {  # don't exceed historic max
                        "constraint_method": "quantile",
                        "constraint_value": 1.0,
                        "constraint_direction": "upper",
                        "constraint_regularization": 1.0,
                        "bounds": True,
                    },
                    {  # don't exceed 2% decline by end of forecast horizon
                        "constraint_method": "slope",
                        "constraint_value": {
                            "slope": -0.02,
                            "window": 28,
                            "window_agg": "min",
                            "threshold": -0.01,
                        },
                        "constraint_direction": "lower",
                        "constraint_regularization": 0.9,
                        "bounds": False,
                    },
                    {  # don't exceed 2% growth by end of forecast horizon
                        "constraint_method": "slope",
                        "constraint_value": {"slope": 0.02, "window": 10, "window_agg": "max", "threshold": 0.01},
                        "constraint_direction": "upper",
                        "constraint_regularization": 0.9,
                        "bounds": False,
                    },
                    {  # don't go below the last 10 values - 10%
                        "constraint_method": "last_window",
                        "constraint_value": {"window": 10, "threshold": -0.1},
                        "constraint_direction": "lower",
                        "constraint_regularization": 1.0,
                        "bounds": False,
                    },
                    {  # don't go below zero
                        "constraint_method": "absolute",
                        "constraint_value": 0,  # can also be an array or Series
                        "constraint_direction": "lower",
                        "constraint_regularization": 1.0,
                        "bounds": True,
                    },
                    {  # don't go below historic min  - 1 st dev
                        "constraint_method": "stdev_min",
                        "constraint_value": 1.0,
                        "constraint_direction": "lower",
                        "constraint_regularization": 1.0,
                        "bounds": True,
                    },
                    {  # don't go above historic mean  + 3 st devs, soft limit
                        "constraint_method": "stdev",
                        "constraint_value": 3.0,
                        "constraint_direction": "upper",
                        "constraint_regularization": 0.5,
                        "bounds": True,
                    },
                    {  # round decimals to 2 places
                        "constraint_method": "round",
                        "constraint_value": 2,
                    },
                    {  # apply dampening (gradually flatten out forecast)
                        "constraint_method": "dampening",
                        "constraint_value": 0.98,
                    },
                ]
            )

        Args:
            constraint_method (str): one of
                stdev_min - threshold is min and max of historic data +/- constraint * st dev of data
                stdev - threshold is the mean of historic data +/- constraint * st dev of data
                absolute - input is array of length series containing the threshold's final value for each
                quantile - constraint is the quantile of historic data to use as threshold
            constraint_regularization (float): 0 to 1
                where 0 means no constraint, 1 is hard threshold cutoff, and in between is penalty term
            upper_constraint (float): or array, depending on method, None if unused
            lower_constraint (float): or array, depending on method, None if unused
            bounds (bool): if True, apply to upper/lower forecast, otherwise False applies only to forecast
            df_train (pd.DataFrame): required for quantile/stdev methods to find threshold values

        Returns:
            self
        """
        self.forecast, self.lower_forecast, self.upper_forecast = apply_constraints(
            self.forecast,
            self.lower_forecast,
            self.upper_forecast,
            constraints=constraints,
            df_train=df_train,
            # old args
            constraint_method=constraint_method,
            constraint_regularization=constraint_regularization,
            upper_constraint=upper_constraint,
            lower_constraint=lower_constraint,
            bounds=bounds,
        )
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  9 22:07:37 2023

@author: colincatlin
"""
import numpy as np


def fourier_extrapolation(
    x, forecast_length=10, n_harm=10, detrend='linear', freq_range=None
):
    m, n = x.shape
    t = np.arange(0, m)

    # Detrend
    if detrend == 'linear':
        p = np.polyfit(t, x, 1).T
        x_notrend = x - np.outer(t, p[:, 0])
    elif detrend == 'quadratic':
        p = np.polyfit(t, x, 2).T
        x_notrend = x - np.outer(t**2, p[:, 0]) - np.outer(t, p[:, 1])
    elif detrend is None:
        x_notrend = x
    else:
        raise ValueError(f"Unsupported detrend option: {detrend}")

    # FFT
    x_freqdom = np.fft.fft(x_notrend, axis=0)

    # Frequencies and sorted indices
    f = np.fft.fftfreq(m)
    indexes = np.argsort(np.abs(f))

    # Frequency range filtering
    if freq_range:
        low, high = freq_range
        indexes = [i for i in indexes if low <= np.abs(f[i]) <= high]

    if n_harm is None:
        use_idx = indexes
    elif isinstance(n_harm, (int, float)):
        # handle float as percentage
        if 0 < n_harm < 1:
            use_idx = indexes[: int(len(indexes) * n_harm)]
        # handle negative percentage ie last N percentage
        elif -1 < n_harm < 0:
            use_idx = indexes[int(len(indexes) * n_harm) :]
        elif n_harm <= -1:
            use_idx = indexes[n_harm * 2 :]
        # handle exact number
        else:
            use_idx = indexes[: 1 + n_harm * 2]
    elif isinstance(n_harm, str):
        if "mid" in n_harm:
            midp = int(''.join(filter(str.isdigit, n_harm)))
            use_idx = indexes[midp : midp + 40]
    else:
        raise ValueError(f"n_harm value {n_harm} not recognized")

    t_extended = np.arange(0, m + forecast_length)
    restored_sig = np.zeros((t_extended.size, n))

    # Use harmonics to reconstruct signal
    for i in use_idx:
        ampli = np.abs(x_freqdom[i]) / m
        phase = np.angle(x_freqdom[i])
        restored_sig += ampli * np.cos(2 * np.pi * f[i] * t_extended[:, None] + phase)
    """
    # Use harmonics to reconstruct signal
    for i in indexes[10:10 + n_harm * 2]:
    # for i in indexes[-2000:]:
        ampli = np.abs(x_freqdom[i]) / m
        phase = np.angle(x_freqdom[i])
        restored_sig += (ampli * np.cos(2 * np.pi * f[i] * t_extended[:, None] + phase))

    nw = pd.DataFrame((restored_sig + np.outer(t_extended, p[:, 0])), columns=df.columns)
    nw.index = df.index.union(pd.date_range(start=df.index[-1], periods=forecast_length+1, freq='D'))
    col = 'FOODS_2_025_TX_1_evaluation'  # 'wiki_all'
    nw['actual'] = df[col]
    nw[['actual', col]].plot()
    """

    # Add trend back
    if detrend == 'linear':
        return restored_sig + np.outer(t_extended, p[:, 0])
    elif detrend == 'quadratic':
        return (
            restored_sig
            + np.outer(t_extended**2, p[:, 0])
            + np.outer(t_extended, p[:, 1])
        )
    else:
        return restored_sig


class FFT(object):
    def __init__(self, n_harm=10, detrend='linear', freq_range=None):
        self.n_harm = n_harm
        self.detrend = detrend
        self.freq_range = freq_range

    def fit(self, x):
        self.m, self.n = x.shape
        t = np.arange(0, self.m)

        # Detrend
        if self.detrend == 'linear':
            self.p = np.polyfit(t, x, 1).T
            x_notrend = x - np.outer(t, self.p[:, 0])
        elif self.detrend == 'quadratic':
            self.p = np.polyfit(t, x, 2).T
            x_notrend = x - np.outer(t**2, self.p[:, 0]) - np.outer(t, self.p[:, 1])
        elif self.detrend == 'cubic':
            self.p = np.polyfit(t, x, 3).T
            x_notrend = (
                x
                - np.outer(t**3, self.p[:, 0])
                - np.outer(t**2, self.p[:, 1])
                - np.outer(t, self.p[:, 2])
            )
        elif self.detrend == 'quartic':
            self.p = np.polyfit(t, x, 4).T
            x_notrend = (
                x
                - np.outer(t**4, self.p[:, 0])
                - np.outer(t**3, self.p[:, 1])
                - np.outer(t**2, self.p[:, 2])
                - np.outer(t, self.p[:, 3])
            )
        elif self.detrend is None:
            x_notrend = x
        else:
            raise ValueError(f"Unsupported detrend option: {self.detrend}")

        # FFT
        self.x_freqdom = np.fft.fft(x_notrend, axis=0)

        # Frequencies and sorted indices
        self.f = np.fft.fftfreq(self.m)
        indexes = np.argsort(np.abs(self.f))

        # Frequency range filtering
        if self.freq_range:
            low, high = self.freq_range
            indexes = [i for i in indexes if low <= np.abs(self.f[i]) <= high]

        if self.n_harm is None:
            use_idx = indexes
        elif isinstance(self.n_harm, (int, float)):
            # handle float as percentage
            if 0 < self.n_harm < 1:
                use_idx = indexes[: int(len(indexes) * self.n_harm)]
            # handle negative percentage ie last N percentage
            elif -1 < self.n_harm < 0:
                use_idx = indexes[int(len(indexes) * self.n_harm) :]
            elif self.n_harm <= -1:
                use_idx = indexes[self.n_harm * 2 :]
            # handle exact number
            else:
                use_idx = indexes[: 1 + self.n_harm * 2]
        elif isinstance(self.n_harm, str):
            if "mid" in self.n_harm:
                midp = int(''.join(filter(str.isdigit, self.n_harm)))
                use_idx = indexes[midp : midp + 41]
        else:
            raise ValueError(f"n_harm value {self.n_harm} not recognized")
        self.use_idx = use_idx

        return self

    def generate_harmonics_dataframe(self, forecast_length=0):
        extended_m = self.m + forecast_length
        harmonics_data = np.zeros((extended_m, len(self.use_idx) * 2))

        for i, idx in enumerate(self.use_idx):
            freq_component = np.fft.ifft(self.x_freqdom[idx], n=self.m, axis=0)
            extended_freq_component = np.tile(
                freq_component, (extended_m // self.m) + 1
            )[:extended_m]
            harmonics_data[:, 2 * i] = np.real(extended_freq_component).flatten()
            harmonics_data[:, 2 * i + 1] = np.imag(extended_freq_component).flatten()

        return harmonics_data

    def predict(self, forecast_length=0):
        # this rather assumes you care only about historical + fcst of length n after
        t_extended = np.arange(0, self.m + forecast_length)
        restored_sig = np.zeros((t_extended.size, self.n))

        # Use harmonics to reconstruct signal
        for i in self.use_idx:
            ampli = np.abs(self.x_freqdom[i]) / self.m
            phase = np.angle(self.x_freqdom[i])
            restored_sig += ampli * np.cos(
                2 * np.pi * self.f[i] * t_extended[:, None] + phase
            )

        # Add trend back
        if self.detrend == 'linear':
            return restored_sig + np.outer(t_extended, self.p[:, 0])
        elif self.detrend == 'quadratic':
            return (
                restored_sig
                + np.outer(t_extended**2, self.p[:, 0])
                + np.outer(t_extended, self.p[:, 1])
            )
        elif self.detrend == 'cubic':
            return (
                restored_sig
                + np.outer(t_extended**3, self.p[:, 0])
                + np.outer(t_extended**2, self.p[:, 1])
                + np.outer(t_extended, self.p[:, 2])
            )
        elif self.detrend == 'quartic':
            return (
                restored_sig
                + np.outer(t_extended**4, self.p[:, 0])
                + np.outer(t_extended**3, self.p[:, 1])
                + np.outer(t_extended**2, self.p[:, 2])
                + np.outer(t_extended, self.p[:, 3])
            )
        else:
            return restored_sig
